package ru.dbotthepony.kstarbound.client.render

import ru.dbotthepony.kstarbound.client.ClientWorld
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.math.FloatMatrix
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ITileChunk
import kotlin.collections.ArrayList

class ChunkRenderer(val state: GLStateTracker, val chunk: Chunk, val world: ClientWorld? = null) : AutoCloseable {
	private inner class TileLayerRenderer(private val layerChangeset: () -> Int, private val isBackground: Boolean) : AutoCloseable {
		private val layers = TileLayerList()
		private val bakedMeshes = ArrayList<BakedStaticMesh>()
		private var changeset = -1

		fun tesselateStatic(view: ITileChunk) {
			if (state.isSameThread()) {
				for (mesh in bakedMeshes) {
					mesh.close()
				}

				bakedMeshes.clear()
			} else {
				unloadableBakedMeshes.addAll(bakedMeshes)
				bakedMeshes.clear()
			}

			layers.clear()

			for ((pos, tile) in view.posToTile) {
				if (tile != null) {
					val renderer = state.tileRenderers.get(tile.def.materialName)
					renderer.tesselate(view, layers, pos, background = isBackground)
				}
			}
		}

		fun loadRenderers(view: ITileChunk) {
			for ((_, tile) in view.posToTile) {
				if (tile != null) {
					state.tileRenderers.get(tile.def.materialName)
				}
			}
		}

		fun uploadStatic(clear: Boolean = true) {
			for ((baked, builder) in layers.buildList()) {
				bakedMeshes.add(BakedStaticMesh(baked, builder))
			}

			if (clear) {
				layers.clear()
			}
		}

		fun render(transform: FloatMatrix<*>) {
			for (mesh in bakedMeshes) {
				mesh.render(transform)
			}
		}

		fun bakeAndRender(transform: FloatMatrix<*>, provider: () -> ITileChunk) {
			if (changeset != layerChangeset.invoke()) {
				this.tesselateStatic(provider.invoke())
				this.uploadStatic()

				changeset = layerChangeset.invoke()
			}

			render(transform)
		}

		override fun close() {
			for (mesh in bakedMeshes) {
				mesh.close()
			}
		}
	}

	private val bakedMeshesForeground = ArrayList<BakedStaticMesh>()
	private val unloadableBakedMeshes = ArrayList<BakedStaticMesh>()

	private val foreground = TileLayerRenderer(chunk.foreground::changeset, isBackground = false)
	private val background = TileLayerRenderer(chunk.background::changeset, isBackground = true)

	private fun getForeground(): ITileChunk {
		return world?.getForegroundView(chunk.pos) ?: chunk.foreground
	}

	private fun getBackground(): ITileChunk {
		return world?.getBackgroundView(chunk.pos) ?: chunk.background
	}

	/**
	 * Тесселирует "статичную" геометрию в builders (к примеру тайлы).
	 *
	 * Может быть вызван вне рендер потока (ибо в любом случае он требует некой "стаитичности" данных в чанке)
	 * но только если до этого был вызыван loadRenderers() и геометрия чанка не поменялась
	 */
	fun tesselateStatic() {
		foreground.tesselateStatic(getForeground())
		background.tesselateStatic(getBackground())
	}

	/**
	 * Принудительно подгружает в GLStateTracker все необходимые рендереры (ибо им нужны текстуры и прочее)
	 *
	 * Вызывается перед tesselateStatic()
	 */
	fun loadRenderers() {
		unloadUnused()

		foreground.loadRenderers(getForeground())
		background.loadRenderers(getBackground())
	}

	private fun unloadUnused() {
		if (unloadableBakedMeshes.size != 0) {
			for (baked in unloadableBakedMeshes) {
				baked.close()
			}

			unloadableBakedMeshes.clear()
		}
	}

	fun uploadStatic(clear: Boolean = true) {
		unloadUnused()

		foreground.uploadStatic(clear)
		background.uploadStatic(clear)
	}

	fun render(transform: FloatMatrix<*> = state.matrixStack.last) {
		unloadUnused()

		background.render(transform)
		foreground.render(transform)
	}

	fun bakeAndRender(transform: FloatMatrix<*> = state.matrixStack.last) {
		unloadUnused()

		background.bakeAndRender(transform, this::getBackground)
		foreground.bakeAndRender(transform, this::getForeground)
	}

	override fun close() {
		background.close()
		foreground.close()
	}
}
